#
# tcl-owy sumulator alg rozpr (uzywa rozszerzenia fiber)
#

# + wersja angielskia, zastapione wyrazy (geany, cale wyrazy), 16.10.2018
# liczbaWierz -> verticesNum
# sasiedzi -> neighbors
# wyslij -> send
# czytaj -> recv
# kom -> msg
# Inicjalizacja -> Initialization
# InicjalizacjaAsynch -> InitializationAsynch
# dostarcz -> deliver
# dostarczKom -> deliverMsg
# licznikKom -> msgCounter
# pokazKom -> showMsg
# runda -> round
# stopien -> degree
# id_los -> id_rand
#

## zmiany:
# + 27.08.2016, glob. tabl. msg, msg(0) = kom0 itd
#

# ---
load ./q3[info sharedlibextension]

## model synchroniczny -----------------------------------------
#

# koniec rundy zaznaczamy przez "fiber yield"

# wierz stopnia np 3 ma incydentne kraw oznaczone przez:
#   kom0, kom1, kom2

# kazda incydentna kraw nr X ma 2 zmienne:
#   komX - zawiera komunikat wyslany przez sasiada w poprzedniej rundzie
#   komX_pisz - tu jest umieszczany komunikat wyslany przez sasiada w biezacej rundzie
#   Uwaga:
#     komX to lista komunikatow, a nie pojedynczy komunikat!!!
#     (jesli komunikat nie ma spacji to lista sie nie rozni
#     od pojedynczego komunikatu!!! to moze byc przyczyna bledow...)

# proc komX_wyslij komunikat
#   sluzy do wysylania komunikatu przez kraw nr X
#   np
#      msg${i}_wyslij "A ku ku"
#      kom1_wyslij $x

# proc komX_dostarcz komunikat (16.11.2009)
#   wysyla i dostarcza komunikat...

# proc send {doKogo co args}
#   wysyla komunikat przez polaczenie nr doKogo
#     np: send $i "A ku ku"
#   podobnie dziala deliver (16.11.2009)

# proc recv {odKogo args}
#   odbiera komunikat z polaczenia nr odKogo
#   w args moze byc opcja -podgraf {lista nr polaczen}
#   komendy send/recv to interf. "wyzszego poziomu"
#     umozliwia pewne uogolnienia, np dzialania na podgrafie:
#       send 0 "123" -podgraf {2 3}; # wysyla przez pol. nr 2
#       recv 1 -podgraf {3 4}; # odczytuje z pol. nr 4
#         !!! jeszcze nie zaimplementowane !!!
#   16.11.2009 - zmiana dzialania:
#     recv pobiera komunikat z komX, jako kolejki komunikatow

# proc. pomocnicze w konsoli:
#   DodajKraw
#   Initialization
#   round
#   set_run
#   fiber_iterate
#   fiber_eval

# proc. pomocnicze w fiberze:
#   iterate zm liIter kod
#   iterate1 zm liIter kod; # robi liIter-1 iteracji od 1 !!!
#   let
#   comment

# BLAD iterate!!! (16.11.2009)
#   problem gdy w ciele iterate jest break/continue/return/error
#   ...chyba usunalem ten blad???

# ---

# proc pomocnicze w zmiennej
# (zeby mozna bylo je latwo przeslac do fiberow)
set iterate_kod {
  proc iterate {zm liIter kod} {
    # catch: 0 ok, 1 error, 2 return, 3 break, 4 continue
    upvar $zm i
    for {set i 0} {$i<$liIter} {incr i} {
      set e [catch {uplevel $kod} x]
      if {$e==3} {return $x}
      if {$e==1 || $e==2} {return -code $e $x}
    }
  }
  proc iterate1 {zm liIter kod} { # ta wersja robi liIter-1 iteracji !!!
    upvar $zm i
    for {set i 1} {$i<$liIter} {incr i} {
      set e [catch {uplevel $kod} x]
      if {$e==3} {return $x}
      if {$e==1 || $e==2} {return -code $e $x}
    }
  }
  proc let {zm wart} {
    uplevel set $zm [expr $wart]
  }
  proc comment x {
  }
  proc unset2 args { # bezpieczne unset; dla zm. globalnych!
    foreach v $args {if {[info exists $v]} {unset $v}}
  }
}
eval $iterate_kod

proc fiber_iterate kod {
  global verticesNum
  iterate i $verticesNum {
    fiber$i eval "uplevel #0 {$kod}"
  }
}
proc fiber_eval {procesor kod} {
  fiber$procesor eval "uplevel #0 {$kod}"
}

# budowanie grafu - proc pomocnicze
proc DodajKraw {v1 v2} {
  global neighbors
  if {[lsearch -exact $neighbors($v1) $v2]==-1} {
    lappend neighbors($v1) $v2
  }
  if {[lsearch -exact $neighbors($v2) $v1]==-1} {
    lappend neighbors($v2) $v1
  }
}

# Initialization fiberow
proc Initialization {} {
  global verticesNum neighbors iterate_kod msgCounter
  set msgCounter 0
  iterate i $verticesNum {
    interp alias fiber$i _puts "" _puts
    interp alias fiber$i incrLicznikKom "" incr ::msgCounter

    # tworzymy komendy komunikacyjne (aliasy) ...
    set degree [llength $neighbors($i)]
    iterate j $degree {
      set x [lindex $neighbors($i) $j]
      set k [lsearch -exact $neighbors($x) $i]
      interp alias fiber$i msg${j}_wyslij fiber$x __wyslij $k
      interp alias fiber$i msg${j}_dostarcz fiber$x __dostarcz $k
    }

    # zmienne dostepne w kazdym fiberze ...
    # + MH, 27.08.2016, dodalem msg($i)
    set id [expr round(rand()*1000)]
    fiber$i eval "
      $iterate_kod
      set run 1
      set id $i
      set id_rand $id
      set degree $degree
      iterate __i $degree {
        upvar #0 msg(\$__i) msg\$__i
        set msg\$__i {}
        set msg\${__i}_pisz {}
      }
      unset __i
      set verticesNum $verticesNum
    "

    # Uwaga: proc msg?_wyslij GWARANTUJE
    #   ze zmodyfikuje zmienne globalne msg?_pisz
    #   a nie przypadkowe zmienne lokalne ...
    fiber$i eval {
      proc __wyslij {nrKom args} {
        global msg${nrKom}_pisz
        eval lappend msg${nrKom}_pisz $args
        incrLicznikKom
      }
      proc __dostarcz {nrKom args} { # "send i deliver"
        global msg${nrKom}
        eval lappend msg${nrKom} $args
        incrLicznikKom
      }
    }

    # proc send/recv (dodane 25.09.2009)
    fiber$i eval {
      proc send {doKogo co args} {
        msg${doKogo}_wyslij $co
      }
      proc deliver {doKogo co args} {
        msg${doKogo}_dostarcz $co
      }
      proc recv {odKogo args} {
        upvar #0 msg$odKogo msg
        set x [lindex $msg 0]
        set msg [lrange $msg 1 end]
        return $x
      }
    }
  }
}

# obsluga zmiennych komunikacyjnych
# te procedure uruchamiac z glownego fibera
proc round {} {
  fiber_iterate {
    iterate __i $degree {
      set msg$__i [set msg${__i}_pisz]
      set msg${__i}_pisz ""
    }
    unset __i
  }
}

# konczenie fiberow
proc set_run p {
  fiber_iterate "set run $p"
}


## model Asynchroniczny ---------------------------------------------
#

## model asynch z osobnymi zdarzeniami obl. i dostarczania (19.10.2009)
# * wysylanie komunikatu przy pomocy:
#     komX_wyslij msg
# * dostarczone komunikaty sa dostepne w zmiennych komX;
#   oczekiwanie na dostarczenie komunikatu okreslnego typu:
#     czytajKomTypu {typ nrPolaczenia}
#     zaklada sie ze komunikat to lista >=2 elementowa, pierwsze slowo to typ
#     jesli brak potrzebnego komunikatu to proc wykonuje "fiber switchto main"
# * generowanie zdarzenia obliczeniowego:
#     fiber switchto procesor
# * generowanie zdarzenia dostarczenia komunikatu:
#     deliverMsg procesor nrPolaczenia
#     przemieszcza komunikaty z komX_pisz do komX
# * podglad dostepnych komunikatow:
#     showMsg
# * proc pomocnicze:
#     zakonczFibery
#     zaproponuj
#       proponuje zdarzenia ktore cos zmienia...

# proc czytajKomTypu {typ nrPolaczenia {varNrPol ""}}
#   zaklada sie ze w kazdym komunikacie pierwsze slowo to typ...
#   mozna podac "*" jako nrPolaczenia i/lub typ !!!
#   rzeczywisty nr jest zwracany przez (opcjonalny) varNrPol
#   (17.11.2009)
#   !!! NIEPRZETESTOWANE i NIEDOKONCZONE !!!
#     niepewne dzialanie zaproponuj...

# ---

proc InitializationAsynch {} {
  Initialization
  fiber_iterate {
    if 0 {
    proc czytajKomTypu {typ nrPolaczenia} {
      global degree id run
      upvar #0 msg$nrPolaczenia msg
      while 1 {
        set i [lsearch $msg "$typ *"]
        if {$i==-1} {
          set ::czytajKomTypu,typ $typ; # wiadomo na co czeka ten fiber
          set ::czytajKomTypu,nrPolaczenia $nrPolaczenia
          fiber switchto main
          set run; # umozliwia zakonczenie pracy
        } else {
          set k [lindex $msg $i]
          set msg [lreplace $msg $i $i]
          return $k
        }
      }
    }}
    if 1 {
    proc czytajKomTypu {typ nrPolaczenia {varNrPol ""}} {
      global degree id run
      while 1 {
        if {$nrPolaczenia!="*"} {
          upvar #0 msg$nrPolaczenia msg; set i [lsearch $msg "$typ *"]
        } else {
          # przeszukujemy wszystkie polaczenia...
          set i -1
          iterate j $degree {
            upvar #0 msg$j msg; set i [lsearch $msg "$typ *"]
            if {$i!=-1} {
              if {$varNrPol!=""} {upvar $varNrPol varNrPol_; set varNrPol_ $j}
              break; # tu byl blad, gdy iterate zle dzialalo...
            }
          }
        }
        if {$i==-1} {
          set ::czytajKomTypu,typ $typ; # wiadomo na co czeka ten fiber!
          set ::czytajKomTypu,nrPolaczenia $nrPolaczenia
	    # obecnie to moze byc "*" !!
          fiber switchto main
          set run; # umozliwia zakonczenie pracy
        } else {
          set k [lindex $msg $i]
          set msg [lreplace $msg $i $i]
          return $k
        }
      }
    }}
  }
}

proc deliverMsg {procesor nrPolaczenia} {
  fiber$procesor eval "uplevel #0 {
    set __nr $nrPolaczenia
  }"
  fiber$procesor eval {uplevel #0 {
    set __k1 [set msg${__nr}_pisz]
    if {[llength $__k1]>0} {
      set __msg [lindex $__k1 0]
      set __k1 [lreplace $__k1 0 0]
      lappend  msg$__nr $__msg
      set msg${__nr}_pisz $__k1
    }
    catch {unset __nr __k1 __msg}
  }}
    # mozna tu uzyc fiber_eval...
}

proc showMsg args {
  fiber_iterate {
    set __wydruk "$id, [format %3d $id_rand]: "
    iterate __i $degree {
      append __wydruk "$__i/ [set msg${__i}_pisz], "
    }
    iterate __i $degree {
      append __wydruk "$__i// [set msg${__i}], "
    }
    _puts $__wydruk
    unset __wydruk __i
  }
}

# zaklada sie ze kazdy fiber ma "set run" za "fiber switchto main"
# dodatkowo proc czysci komunikaty!
proc zakonczFibery {} {
  global verticesNum
  set err [fiber error]
  for {set i 0} {$i<$verticesNum} {incr i} {
    catch {fiber_eval $i {unset run}}
    catch {fiber switchto $i}
    catch {fiber_eval $i {set run 1}}
  }
  fiber_iterate {
    iterate __i $degree {
      set msg$__i {}
      set msg${__i}_pisz {}
    }
    unset __i
  }
  set ::msgCounter 0
}

proc zaproponuj {} {
  fiber_iterate {
    proc __zaproponuj {} {
      upvar ::czytajKomTypu,typ typ
      upvar ::czytajKomTypu,nrPolaczenia nr
      global id degree
      if {! ([info exists typ] && [info exists nr])} {return ""}
      if {$nr=="*"} {
        iterate i $degree {
          upvar ::msg${i}_pisz kom_pisz
  	  if {[lindex $kom_pisz 0 0]==$typ || ($typ=="*" && $kom_pisz!={}) } {
	    return "deliverMsg $id $i; fiber switchto $id"
	  }
	}
      } else {
        upvar ::msg${nr}_pisz kom_pisz
	if {[lindex $kom_pisz 0 0]==$typ || ($typ=="*" && $kom_pisz!={}) } {
	  return "deliverMsg $id $nr; fiber switchto $id"
	}
      }
      return ""
    }
  }
  global verticesNum
  set xx "\n"
  iterate i $verticesNum {
    set y [fiber_eval $i __zaproponuj]
    if {$y!=""} {append xx "$y\n"}
  }
  return [string range $xx 0 end-1]
}

